package com.solibri.smc.api.examples.beginner;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import com.solibri.geometry.linearalgebra.Vector3d;
import com.solibri.geometry.mesh.TriangleMesh;
import com.solibri.geometry.primitive3d.Triangle3d;
import com.solibri.smc.api.SMC;
import com.solibri.smc.api.checking.DoubleParameter;
import com.solibri.smc.api.checking.FilterParameter;
import com.solibri.smc.api.checking.OneByOneRule;
import com.solibri.smc.api.checking.Result;
import com.solibri.smc.api.checking.ResultFactory;
import com.solibri.smc.api.checking.RuleParameters;
import com.solibri.smc.api.distance.Distance;
import com.solibri.smc.api.filter.AABBIntersectionFilter;
import com.solibri.smc.api.filter.ComponentFilter;
import com.solibri.smc.api.model.Component;
import com.solibri.smc.api.model.PropertyType;
import com.solibri.smc.api.ui.UIContainer;
import com.solibri.smc.api.visualization.VisualizationItem;

/**
* An example check that visualizes all the max distances of all component pairs in a
* component set; each max distance comes from a pair of components in the component set.
*
* A max distance is calculated using the maximum distance between all the points
* in a component pair.
 */
public final class DistanceVisualizationRule extends OneByOneRule {

	/**
	 * Retrieve the parameter creation handler, used to define parameters for
	 * this rule.
	 */
	private final RuleParameters params = RuleParameters.of(this);

	/**
	 * Retrieve the default filter.
	 * Every component that passes the filter is then forwarded to
	 * the {@link OneByOneRule#check(Component, ResultFactory)} method.
	 */
	final FilterParameter rpComponentFilter = this.getDefaultFilterParameter();

	final FilterParameter rpComponentFilter2 = params.createFilter("rpComponentFilter2");

	final DoubleParameter maximumDistance = params.createDouble("rpMaximumDistance", PropertyType.LENGTH);

	/**
	 * Add the UI definition from class DistanceVisualizationRuleUIDefinition.
	 */
	private final DistanceVisualizationRuleUIDefinition uiDefinition = new DistanceVisualizationRuleUIDefinition(this);

	@Override
	public Collection<Result> check(Component component, ResultFactory resultFactory) {

		// Search only for components that are in range of the maximumDistance parameter.
		double searchDistance = maximumDistance.getValue();
		ComponentFilter filter2 = rpComponentFilter2.getValue();
		ComponentFilter withinDistanceFilter =
			AABBIntersectionFilter.ofComponentBounds(component, searchDistance, searchDistance).and(filter2);
		Collection<Component> targets = SMC.getModel().getComponents(withinDistanceFilter);

		List<Result> results = new ArrayList<>();

		// Check each component separately.
		for (Component target : targets) {
			results.add(distanceCheck(component, target, resultFactory));
		}

		return results;
	}

	@Override
	public UIContainer getParametersUIDefinition() {
		return uiDefinition.getDefinitionContainer();
	}

	/**
	 * Calculates the maximum distance between two components.
	 *
	 * @param component1 the first component
	 * @param component2 the second component
	 * @return the maximum distance between any points of components
	 */
	private Distance maxDistance(Component component1, Component component2) {
		/*
		 * The triangles in component's geometry can be looped with the TriangleMesh used as an Iterable of triangles.
		 */
		TriangleMesh sourceMesh = component1.getTriangleMesh();
		TriangleMesh targetMesh = component2.getTriangleMesh();

		Distance maxDistance = null;

		for (Triangle3d sourceTriangle : sourceMesh) {
			for (Triangle3d targetTriangle : targetMesh) {
				Distance potentialDistance = maxDistance(sourceTriangle, targetTriangle);
				if (maxDistance == null || potentialDistance.getLength() > maxDistance.getLength()) {
					maxDistance = potentialDistance;
				}
			}
		}

		return maxDistance;
	}

	/**
	 * Calculates the maximum distance between two triangles.
	 * @param triangle1 the first triangle
	 * @param triangle2 the second triangle
	 * @return the maximum distance between any points of two triangles
	 */
	private Distance maxDistance(Triangle3d triangle1, Triangle3d triangle2) {
		Distance maxDistance = null;
		maxDistance = updateDistance(triangle1.getFirstPoint(), triangle2.getFirstPoint(), maxDistance);
		maxDistance = updateDistance(triangle1.getFirstPoint(), triangle2.getSecondPoint(), maxDistance);
		maxDistance = updateDistance(triangle1.getFirstPoint(), triangle2.getThirdPoint(), maxDistance);

		maxDistance = updateDistance(triangle1.getSecondPoint(), triangle2.getFirstPoint(), maxDistance);
		maxDistance = updateDistance(triangle1.getSecondPoint(), triangle2.getSecondPoint(), maxDistance);
		maxDistance = updateDistance(triangle1.getSecondPoint(), triangle2.getThirdPoint(), maxDistance);

		maxDistance = updateDistance(triangle1.getThirdPoint(), triangle2.getFirstPoint(), maxDistance);
		maxDistance = updateDistance(triangle1.getThirdPoint(), triangle2.getSecondPoint(), maxDistance);
		maxDistance = updateDistance(triangle1.getThirdPoint(), triangle2.getThirdPoint(), maxDistance);

		return maxDistance;
	}

	/**
	 * A helper method for updating the Distance object for a given pair of points.
	 *
	 * @param firstPoint the first point
	 * @param secondPoint the second points
	 * @param maxDistance the current Distance instance
	 * @return the new Distance object
	 */
	private static Distance updateDistance(Vector3d firstPoint, Vector3d secondPoint, Distance maxDistance) {
		double potentialDistance = firstPoint.distance(secondPoint);
		if (maxDistance == null || potentialDistance > maxDistance.getLength()) {
			maxDistance = new Distance(potentialDistance, firstPoint, secondPoint);
		}
		return maxDistance;
	}

	private Result distanceCheck(Component source, Component target, ResultFactory results) {
		// Calculate the distance.
		Distance distance = maxDistance(source, target);

		String nameSource = source.getName();
		String nameTarget = target.getName();
		String distanceText = PropertyType.LENGTH.getFormat().format(distance);

		// Create the result.
		return results
			.create(
				nameSource + " " + nameTarget + " " + distanceText,
				"Distance between " + nameSource + " " + nameTarget + " is " + distanceText)
			.withInvolvedComponent(target)
			.withVisualization(visualization -> {
				// Create visualization using provided factory that can create dimensional lines.
				List<VisualizationItem> dimensionLine = VisualizationItem
					.createDimension(distance.getStartPoint(), distance.getEndPoint());
				visualization.addVisualizationItems(dimensionLine);
				visualization.addComponent(source, 0.5);
				visualization.addComponent(target, 0.5);
			});
	}

}
